package heap

import (
	"474420502.top/eson/structure/compare"
)

type Heap struct {
	size     int
	elements *arrayN
	Compare  compare.Compare
}

func New(Compare compare.Compare) *Heap {
	h := &Heap{Compare: Compare}
	h.elements = newWithCap(8, 8, 8)
	return h
}

func (h *Heap) Values() []interface{} {
	return h.elements.Values()[0:h.size]
}

func (h *Heap) Push(v interface{}) {

	if h.size >= h.elements.Cap() {
		h.elements.Grow(h.elements.zsize * 2)
	}

	curidx := h.size
	curarray, curarrayidx := h.elements.getOfSet(curidx, v)

	h.size++

	// up
	for curidx != 0 {
		pidx := (curidx - 1) / 2
		parray, parrayidx := h.elements.get(pidx)
		pvalue := parray[parrayidx]
		if h.Compare(v, pvalue) > 0 {
			curarray[curarrayidx], parray[parrayidx] = parray[parrayidx], curarray[curarrayidx]
			curidx = pidx
			curarray = parray
			curarrayidx = parrayidx
		} else {
			break
		}
	}
}

func (h *Heap) Pop() (interface{}, bool) {

	if h.size == 0 {
		return nil, false
	}

	h.size--

	curidx := 0
	curarray, curarrayidx := h.elements.get(curidx)
	curvalue := curarray[curarrayidx]
	if h.size == 0 {
		return curvalue, true
	}

	result := curvalue

	lastarray, lastarrayidx := h.elements.get(h.size)
	curarray[curarrayidx] = lastarray[lastarrayidx]
	curvalue = curarray[curarrayidx]

	var childidx, childarrayidx int
	var childarray []interface{}
	var childvalue interface{}
	// down
	for curidx < h.size {

		var childidx1, childidx2, childarrayidx1, childarrayidx2 int
		var childarray1, childarray2 []interface{}
		var childvalue1, childvalue2 interface{}

		movidx := curidx << 1
		childidx2 = movidx + 2
		if childidx2 < h.size {
			childarray2, childarrayidx2 = h.elements.get(childidx2)
			childvalue2 = childarray2[childarrayidx2]

			childidx1 = curidx<<1 + 1
			childarray1, childarrayidx1 = h.elements.get(childidx1)
			childvalue1 = childarray1[childarrayidx1]

			if h.Compare(childvalue1, childvalue2) >= 0 {
				childvalue = childvalue1
				childidx = childidx1
				childarray = childarray1
				childarrayidx = childarrayidx1
			} else {
				childvalue = childvalue2
				childidx = childidx2
				childarray = childarray2
				childarrayidx = childarrayidx2
			}

		} else {

			childidx1 = movidx + 1
			if childidx1 < h.size {
				childarray1, childarrayidx1 = h.elements.get(childidx1)
				childvalue1 = childarray1[childarrayidx1]

				childvalue = childvalue1
				childidx = childidx1
				childarray = childarray1
				childarrayidx = childarrayidx1
			} else {
				break
			}
		}

		if h.Compare(curvalue, childvalue) < 0 {
			curarray[curarrayidx] = childvalue
			childarray[childarrayidx] = curvalue
			curidx = childidx
			curarray = childarray
			curarrayidx = childarrayidx
		} else {
			break
		}
	}

	return result, true

}

func (h *Heap) shiftdown(v interface{}) {

}

func (h *Heap) shiftup(v interface{}) {

}
